// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
//
// Sample chicagoGLWidget
// --------------------------------------
// Copyright (C) 2006-2008 Nicolas Schulz
//
//
// This sample source file is not covered by the LGPL as the rest of the SDK
// and may be used without any restrictions
//
// *************************************************************************************************

#include <math.h>
#include <iomanip>
#include <QTimer>

#include "chicagoGLWidget.h"
#include "Horde3D.h"
#include "Horde3DUtils.h"

// Convert from degrees to radians
inline float degToRad( float f ) 
{
	return f * (3.1415926f / 180.0f);
}


chicagoGLWidget::chicagoGLWidget()
{
	for( unsigned int i = 0; i < 320; ++i ) _keys[i] = false;

	_x = 15; 
    _y = 3; 
    _z = 20; 
    _rx = -10; 
    _ry = 60; 
    _velocity = 10.0f;
	_curFPS = 30; 
    _timer = 0;

	_freeze = false; 
    _showFPS = false; 
    _debugViewMode = false; 
    _wireframeMode = false;
	_cam = 0;
}

chicagoGLWidget::~chicagoGLWidget()
{
	delete _crowdSim; _crowdSim = 0x0;
	
	// Release engine
	Horde3D::release();
}

void chicagoGLWidget::animate()
{
    QTimer::singleShot(45, this, SLOT(animate()));
    updateGL();
    emit animated();
}

QSize chicagoGLWidget::sizeHint() const
{
    return QSize(800,600);
}

QSize chicagoGLWidget::minimumSizeHint() const
{
    return QSize(800,600);
}

void chicagoGLWidget::initializeGL()
{	
	// Initialize engine
	if( !Horde3D::init() )
	{	
		Horde3DUtils::dumpMessages();
		return;
	}
	
	// Set paths for resources
	Horde3DUtils::setResourcePath( ResourceTypes::SceneGraph, "models" );
	Horde3DUtils::setResourcePath( ResourceTypes::Geometry, "models" );
	Horde3DUtils::setResourcePath( ResourceTypes::Animation, "models" );
	Horde3DUtils::setResourcePath( ResourceTypes::Material, "materials" );
	Horde3DUtils::setResourcePath( ResourceTypes::Code, "shaders" );
	Horde3DUtils::setResourcePath( ResourceTypes::Shader, "shaders" );
	Horde3DUtils::setResourcePath( ResourceTypes::Texture2D, "textures" );
	Horde3DUtils::setResourcePath( ResourceTypes::TextureCube, "textures" );
	Horde3DUtils::setResourcePath( ResourceTypes::Effect, "effects" );
	Horde3DUtils::setResourcePath( ResourceTypes::Pipeline, "pipelines" );

	// Set options
	Horde3D::setOption( EngineOptions::LoadTextures, 1 );
	Horde3D::setOption( EngineOptions::TexCompression, 0 );
	Horde3D::setOption( EngineOptions::AnisotropyFactor, 8 );
	Horde3D::setOption( EngineOptions::ShadowMapSize, 2048 );
	Horde3D::setOption( EngineOptions::FastAnimation, 1 );

	Horde3D::setOption( EngineOptions::DebugViewMode, _debugViewMode ? 1.0f : 0.0f );
	Horde3D::setOption( EngineOptions::WireframeMode, _wireframeMode ? 1.0f : 0.0f );

	// Add resources
	// Pipelines
	_forwardPipeRes = Horde3D::addResource( ResourceTypes::Pipeline, "forward.pipeline.xml", 0 );
	_deferredPipeRes = Horde3D::addResource( ResourceTypes::Pipeline, "deferred.pipeline.xml", 0 );
	// Font
	_fontMatRes = Horde3D::addResource( ResourceTypes::Material, "font.material.xml", 0 );
	// Logo
	_logoMatRes = Horde3D::addResource( ResourceTypes::Material, "logo.material.xml", 0 );
	// Shader for deferred shading
	ResHandle lightMatRes = Horde3D::addResource( ResourceTypes::Material, "light.material.xml", 0 );
	// Environment
	ResHandle envRes = Horde3D::addResource( ResourceTypes::SceneGraph, "platform.scene.xml", 0 );
	// Skybox
	ResHandle skyBoxRes = Horde3D::addResource( ResourceTypes::SceneGraph, "skybox.scene.xml", 0 );
	
	// Load resources
	Horde3DUtils::loadResourcesFromDisk( "../Content" );

	// Add scene nodes
	// Add camera
	_cam = Horde3D::addCameraNode( RootNode, "Camera", _forwardPipeRes );
	//Horde3D::setNodeParami( _cam, CameraNodeParams::OcclusionCulling, 1 );
	// Add environment
	NodeHandle env = Horde3D::addNodes( RootNode, envRes );
	Horde3D::setNodeTransform( env, 0, 0, 0, 0, 0, 0, 0.23f, 0.23f, 0.23f );
	// Add skybox
	NodeHandle sky = Horde3D::addNodes( RootNode, skyBoxRes );
	Horde3D::setNodeTransform( sky, 0, 0, 0, 0, 0, 0, 210, 50, 210 );
	// Add light source
	NodeHandle light = Horde3D::addLightNode( RootNode, "Light1", lightMatRes, "LIGHTING", "SHADOWMAP" );
	Horde3D::setNodeTransform( light, 0, 20, 50, -30, 0, 0, 1, 1, 1 );
	Horde3D::setNodeParamf( light, LightNodeParams::Radius, 200 );
	Horde3D::setNodeParamf( light, LightNodeParams::FOV, 90 );
	Horde3D::setNodeParami( light, LightNodeParams::ShadowMapCount, 3 );
	Horde3D::setNodeParamf( light, LightNodeParams::ShadowSplitLambda, 0.9f );
	Horde3D::setNodeParamf( light, LightNodeParams::ShadowMapBias, 0.001f );
	Horde3D::setNodeParamf( light, LightNodeParams::Col_R, 0.9f );
	Horde3D::setNodeParamf( light, LightNodeParams::Col_G, 0.7f );
	Horde3D::setNodeParamf( light, LightNodeParams::Col_B, 0.75f );

	_crowdSim = new CrowdSim();
	_crowdSim->init();
}

void chicagoGLWidget::resizeGL(int width, int height)
{
    Horde3D::resize(0,0,width,height);
    // Horde3D::resize(0,0,800,600);
    // Horde3D::setupCameraView(cam,45.0f,(float)width /height,0.1f,1000.0f);
}

void chicagoGLWidget::paintGL()
{
    _timer += 1 / _curFPS;

    if( !_freeze )
    {
        _crowdSim->update( _curFPS );
    }

    // Set camera parameters
    Horde3D::setNodeTransform( _cam, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );

    if( _showFPS )
    {
        // Avoid updating FPS text every frame to make it readable
        if( _timer > 0.3f )
        {
            _fpsText.str( "" );
            _fpsText << "FPS: " << fixed << setprecision( 2 ) << _curFPS;
            _timer = 0;
        }

        // Show text
        Horde3DUtils::showText( _fpsText.str().c_str(), 0, 0.95f, 0.03f, 0, _fontMatRes );
    }

    // Show logo
    Horde3D::showOverlay( 0.75f, 0, 0, 0, 1, 0, 1, 0,
            1, 0.2f, 1, 1, 0.75f, 0.2f, 0, 1,
            7, _logoMatRes );

    // Render scene
    Horde3D::render( _cam );

    // Remove all overlays
    Horde3D::clearOverlays();

    // Write all mesages to log file
    Horde3DUtils::dumpMessages();
}

void chicagoGLWidget::keyPressEvent(QKeyEvent *event)
{
    switch(event->key())
    {
        case Qt::Key_Space:
            _freeze = !_freeze;
            break;

        case Qt::Key_F1:
            showFullScreen();
            break;

        case Qt::Key_F2:
        case Qt::Key_Escape:
            showNormal();
            break;

        case Qt::Key_F3:
            {
                if( Horde3D::getNodeParami( _cam, CameraNodeParams::PipelineRes ) == _forwardPipeRes )
                    Horde3D::setNodeParami( _cam, CameraNodeParams::PipelineRes, _deferredPipeRes );
                else
                    Horde3D::setNodeParami( _cam, CameraNodeParams::PipelineRes, _forwardPipeRes );
            }
            break;

        case Qt::Key_F7:
            _debugViewMode = !_debugViewMode;
            break;

        case Qt::Key_F8:
            _wireframeMode = !_wireframeMode;
            break;

        case Qt::Key_F9:
            _showFPS = !_showFPS;
            break;

        case Qt::Key_W:
        case Qt::Key_S:
        case Qt::Key_A:
        case Qt::Key_D:
            {
                float curVel = _velocity / _curFPS;

                if(Qt::SHIFT==event->modifiers())
                    curVel *= .5;

                int key=event->key();
                if (key==Qt::Key_W)
                {
                    // Move forward
                    _x -= sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
                    _y -= sinf( -degToRad( _rx ) ) * curVel;
                    _z -= cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
                }
                else if(key==Qt::Key_S)
                {
                    // Move backward
                    _x += sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
                    _y += sinf( -degToRad( _rx ) ) * curVel;
                    _z += cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
                }
                else if(key==Qt::Key_A)
                {
                    // Strafe left
                    _x += sinf( degToRad( _ry - 90) ) * curVel;
                    _z += cosf( degToRad( _ry - 90 ) ) * curVel;
                }
                else if(key==Qt::Key_D)
                {
                    // Strafe right
                    _x += sinf( degToRad( _ry + 90 ) ) * curVel;
                    _z += cosf( degToRad( _ry + 90 ) ) * curVel;
                }
            }
    }
}

void chicagoGLWidget::mousePressEvent(QMouseEvent *event)
{
    lastPos = event->pos();
}

void chicagoGLWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - lastPos.x();
    int dy = event->y() - lastPos.y();

    lastPos = event->pos();

    _ry -= dx;
    _rx -= dy;

    // if( _rx > 90 ) 
    //     _rx = 90; 
    // if( _rx < -90 ) 
    //     _rx = -90;
}


